#include <WiFiClient.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h>

const char* ssid = "";
const char* password =  "";

const char* mqttClient =  "ESP8266 - Valves";
const char* mqttServer = "";
const int mqttPort = 1883;
//const char* mqttUser = "YourMQTTUsername";
//const char* mqttPassword = "YourMQTTPassword";

const char* BUZZER_TOPIC = "/home/hall/buzzer";
const char* VOLTAGE_TOPIC = "/home/toilet/esp/voltage";
const char* VOLTAGE_TOPIC_ACP = "/home/toilet/esp/voltage/acp";
/*  Затопление. 1/0. При затоплении закрываем кран, включаем BUZZER, отправляем тревогу на брокер*/
const char* LEAK_TOPIC = "/home/toilet/leak";
/* Усправляет кранами горячей и холодной воды */
const char* VALVE_TOPIC = "/home/toilet/valve/";

const int VALVE_COUNT = 1;
// усредняем по 60 показаниям, раз в минуту
const int MEASURE_AVERAGING = 60;
float chargeAvg[MEASURE_AVERAGING];
int chargeStep = 0; // счетчик шага усреднения показания
/* хранение значений сенсоров, чтобы не отправлять повторно топик с тем же значением.*/
bool leak = false;
//считаем вентили включенными по дефолту
bool valves[VALVE_COUNT] = {true};

/* PINS */
const int CHARGE_PIN = A0; // select the input pin for LDR
const int LEAK_PIN = D5;
const int BUZZER_PIN = D4;
/* По 2 пина на каждый кран. 1 пин - закрытие, 2 пин - открытие */
const int VALVE_PINS[VALVE_COUNT * 2] = {D0, D1};

WiFiClient espClient;
PubSubClient client(espClient);


void setup()
{
  Serial.begin(115200);
  Serial.println("Started...");

  // watchdog
  ESP.wdtDisable();
  ESP.wdtEnable(WDTO_8S);
  
  Serial.println("Set MQTT...");
  client.setServer(mqttServer, mqttPort);
  client.setCallback(callback);

  Serial.println("Set pin modes...");
  pinMode(LEAK_PIN, INPUT);
  pinMode(BUZZER_PIN, OUTPUT);

  for (int i  = 0; i < VALVE_COUNT * 2; i++) {
    pinMode(VALVE_PINS[i], OUTPUT);
    // отключаем релешки
    digitalWrite(VALVE_PINS[i], 1);
  }
  for (int i  = 1; i <= VALVE_COUNT; i++) {
    valves[i-1] = false;
    setValveMode(i, false);
  }
}

void loop()
{
  ESP.wdtFeed();
  // подключаемся к wi-fi
  if (WiFi.status() != WL_CONNECTED) {
    if (!connectWifi()) return;
  }

  // подключаемся к MQTT серверу
  if (WiFi.status() == WL_CONNECTED) {
    if (!client.connected()) {
      if (!connectMqtt()) return;
      clientInitTopics();
    }
    client.loop();
    collectSensorData();
    delay(1000);
  }
}

/* Получение события по MQTT */
void callback(char* topic, byte* payload, unsigned int length)
{
  String value = "";
  Serial.print("MQTT message arrived in topic \"");
  Serial.print(topic);
  Serial.print("\". ");
  Serial.print("Message: \"");
  for (int i = 0; i < length; i++) {
    value += (char)payload[i];
  }
  int newPos = value.toInt(); // 0 or 1
  Serial.print(newPos);
  Serial.println("\"");

  if (String(topic) == String(BUZZER_TOPIC)) {
    enableBuzzer(newPos == 1);
    return;
  }

  for (int i  = 1; i <= VALVE_COUNT; i++) {
    String relayTopic = String(VALVE_TOPIC) + String(i);
    if (String(topic) == String(relayTopic)) {
      bool enable = newPos == 1;
      // исключаем обработку дублирующих сообщений из брокера.
      // сам метод закрытия оставляем чистым для проверок
      if (valves[i - 1] == enable) return;
      valves[i - 1] = enable;
      setValveMode(i, enable);
      return;
    }
  }
}

/* Подключение к сети */
bool connectWifi() {
  Serial.print("Wifi status: ");
  Serial.println(WiFi.status());
  Serial.print("Connecting to ");
  Serial.print(ssid);
  Serial.println("...");
  WiFi.begin(ssid, password);

  if (WiFi.waitForConnectResult() != WL_CONNECTED) {
    delay(2000);
    return false;
  }
  Serial.print("Connected to the WiFi network - ");
  Serial.println(WiFi.localIP());
  return true;
}

/* Подключение к серверу MQTT */
bool connectMqtt() {
  Serial.print("Client state ");
  Serial.println(client.state());

  Serial.print("Connecting to MQTT server ");
  Serial.print(mqttServer);
  Serial.println("...");

  //      if (client.connect(mqttClient, mqttUser, mqttPassword))
  if (!client.connect(mqttClient))
  {
    Serial.print("Could not connect to MQTT server. Failed with state ");
    Serial.println(client.state());
    delay(2000);
    return false;
  }
  Serial.println("MQTT connected");
  return true;
}

/* Подписка на события */
void clientInitTopics() {
  for (int i  = 1; i <= VALVE_COUNT; i++) {
    String relayTopic = String(VALVE_TOPIC) + String(i);
    clientSubscribeTopic(relayTopic.c_str());
  }

  clientSubscribeTopic(BUZZER_TOPIC);
}

/* Подписка на топик */
void clientSubscribeTopic(const char* topic) {
  Serial.print("MQTT subscribe topic \"");
  Serial.print(topic);
  Serial.println("\"");
  client.subscribe(topic);
}

/* Отправка на топик */
void clientSendTopic(const char* topic, const char* value) {
  Serial.print("MQTT publish topic \"");
  Serial.print(topic);
  Serial.print("\"");
  Serial.print(" with value \"");
  Serial.print(value);
  Serial.println("\"");
  client.publish(topic, value);
}


/* Функция отправки показаний с датчика */
void collectSensorData() {
  delay(250);
  getLeakState();
  getCharge();
}


void setValveMode(int valveNumber, bool enable) {
  Serial.print("Set valve ");
  Serial.print(valveNumber);
  Serial.println(enable ? " enabled: " : " disabled: ");

  int closePin = VALVE_PINS[valveNumber * 2 - 2];
  int openPin = VALVE_PINS[valveNumber * 2 - 1];

  // открытие
  if (enable) {
    // включить открытие
    Serial.print("Enable pin " );
    Serial.println(openPin);
    digitalWrite(openPin, 0); // invert relay
  }
  delay(100);

  // закрытие
  if (!enable) {
    // включить закрытие
    Serial.print("Enable pin " );
    Serial.println(closePin);
    digitalWrite(closePin, 0); // invert relay
  }

  // чтобы успел закрыться
  delay(10000);

  // отключаем релешки
  Serial.print("Disable pin " );
  Serial.print(openPin);
  Serial.print(", " );
  Serial.println(closePin);
  digitalWrite(openPin, 1);
  digitalWrite(closePin, 1);
}

void enableBuzzer(bool enable) {
  Serial.print("Set buzzer " );
  Serial.println(enable );

  if (enable) {
    tone(BUZZER_PIN, 800); // Send 1KHz sound signal...
  } else {
    noTone(BUZZER_PIN);     // Stop sound...
  }
}

///* Напряжение батареи устройства % */
//void getChargePersent() {
//  if (chargeStep < MEASURE_AVERAGING) {
//    float pos = analogRead(CHARGE_PIN);
//    chargeAvg[chargeStep] = pos;
//    chargeStep++;
//    return;
//  }
//  chargeStep = 0;
//
//  float charge = average(chargeAvg, MEASURE_AVERAGING) ; // среднее показание
//  float chargePercent = (charge / 1024.0) * 100.0; //persent
//  String sensorValue = String(chargePercent);
//  clientSendTopic(VOLTAGE_TOPIC, sensorValue.c_str() );
//}


// сколько позиций АЦП в 1 вольте
const float oneVolt = 70;
/* Напряжение батареи */
void getCharge() {
  if (chargeStep < MEASURE_AVERAGING) {
    float pos = analogRead(CHARGE_PIN);
    chargeAvg[chargeStep] = pos;
    chargeStep++;
    return;
  }
  chargeStep = 0;

  float charge = average(chargeAvg, MEASURE_AVERAGING) ; // среднее показание
  float chargeVolts = charge / oneVolt; //volts
  String sensorValue = String(chargeVolts);
  clientSendTopic(VOLTAGE_TOPIC, sensorValue.c_str() );
  
  String sensorValueAcp = String(charge);
  clientSendTopic(VOLTAGE_TOPIC_ACP, sensorValueAcp.c_str() );
}

/* Протечка в помещении */
void getLeakState() {
  bool l = digitalRead(LEAK_PIN) == 0; //invert
  if (l == leak) return;

  leak = l;
  String sensorValue = String(leak ? 1 : 0);
  Serial.print("Room leak " );
  Serial.println(sensorValue );
  clientSendTopic(LEAK_TOPIC, sensorValue.c_str() );

  clientSendTopic(BUZZER_TOPIC, sensorValue.c_str() );
  enableBuzzer(leak);

  // автоматическое закрытие, ручное открытие!
  if (leak) {
    for (int i  = 1; i <= VALVE_COUNT; i++) {
      String relayTopic = String(VALVE_TOPIC) + String(i);
      clientSendTopic(relayTopic.c_str(), "0");
      setValveMode(i, false);
    }
  }
}

/* UTILS */

float average(float * array, int len)  // assuming array is float.
{
  float sum = 0;  // sum will be larger than an item, long for safety.
  for (int i = 0 ; i < len ; i++)
    sum += array[i] ;
  return  sum / (float)len ;  // average will be fractional, so float may be appropriate.
}


//int state ()
//Returns the current state of the client. If a connection attempt fails, this can be used to get more information about the failure.
//Returns
//
//    int - the client state, which can take the following values (constants defined in PubSubClient.h):
//        -4 : MQTT_CONNECTION_TIMEOUT - the server didn't respond within the keepalive time
//        -3 : MQTT_CONNECTION_LOST - the network connection was broken
//        -2 : MQTT_CONNECT_FAILED - the network connection failed
//        -1 : MQTT_DISCONNECTED - the client is disconnected cleanly
//        0 : MQTT_CONNECTED - the client is connected
//        1 : MQTT_CONNECT_BAD_PROTOCOL - the server doesn't support the requested version of MQTT
//        2 : MQTT_CONNECT_BAD_CLIENT_ID - the server rejected the client identifier
//        3 : MQTT_CONNECT_UNAVAILABLE - the server was unable to accept the connection
//        4 : MQTT_CONNECT_BAD_CREDENTIALS - the username/password were rejected
//        5 : MQTT_CONNECT_UNAUTHORIZED - the client was not authorized to connect


